import copy

from . import chinese, chinese_mix, cleaned_text_to_sequence, english, french, japanese, korean, spanish, thai, turkish

language_module_map = {
    "ZH": chinese,
    "JP": japanese,
    "EN": english,
    'ZH_MIX_EN': chinese_mix,
    'KR': korean,
    'FR': french,
    'SP': spanish,
    'ES': spanish,
    'TH': thai,
    'TR': turkish
}


def clean_text(text, language):
    language_module = language_module_map[language]
    norm_text = language_module.text_normalize(text)
    phones, tones, word2ph = language_module.g2p(norm_text)
    return norm_text, phones, tones, word2ph


def clean_text_bert(text, language, device=None):
    language_module = language_module_map[language]
    norm_text = language_module.text_normalize(text)
    phones, tones, word2ph = language_module.g2p(norm_text)

    word2ph_bak = copy.deepcopy(word2ph)
    for i in range(len(word2ph)):
        word2ph[i] = word2ph[i] * 2
    word2ph[0] += 1
    bert = language_module.get_bert_feature(norm_text, word2ph, device=device)

    return norm_text, phones, tones, word2ph_bak, bert


def text_to_sequence(text, language):
    norm_text, phones, tones, word2ph = clean_text(text, language)
    return cleaned_text_to_sequence(phones, tones, language)


if __name__ == "__main__":
    pass



import argparse
import glob
import json
import logging
import os
import subprocess

import librosa
import numpy as np
import torch
import torchaudio
from scipy.io.wavfile import read

from meloplus import commons
from meloplus.text import cleaned_text_to_sequence, get_bert
from meloplus.text.cleaner import clean_text

MATPLOTLIB_FLAG = False

logger = logging.getLogger(__name__)


def get_text_for_tts_infer(text, language_str, hps, device, symbol_to_id=None):
    norm_text, phone, tone, word2ph = clean_text(text, language_str)
    phone, tone, language = cleaned_text_to_sequence(phone, tone, language_str, symbol_to_id)

    if hps.data.add_blank:
        phone = commons.intersperse(phone, 0)
        tone = commons.intersperse(tone, 0)
        language = commons.intersperse(language, 0)
        for i in range(len(word2ph)):
            word2ph[i] = word2ph[i] * 2
        word2ph[0] += 1

    if getattr(hps.data, "disable_bert", False):
        bert = torch.zeros(1024, len(phone))
        ja_bert = torch.zeros(768, len(phone))
    else:
        bert = get_bert(norm_text, word2ph, language_str, device)
        del word2ph
        assert bert.shape[-1] == len(phone), phone

        if language_str == "ZH":
            bert = bert
            ja_bert = torch.zeros(768, len(phone))
        elif language_str in ["JP", "EN", "ZH_MIX_EN", 'KR', 'SP', 'ES', 'FR', 'DE', 'RU', 'TH', 'TR']:
            ja_bert = bert
            bert = torch.zeros(1024, len(phone))
        else:
            raise NotImplementedError()

    assert bert.shape[-1] == len(phone), f"Bert seq len {bert.shape[-1]} != {len(phone)}"

    phone = torch.LongTensor(phone)
    tone = torch.LongTensor(tone)
    language = torch.LongTensor(language)
    return bert, ja_bert, phone, tone, language


def load_checkpoint(checkpoint_path, model, optimizer=None, skip_optimizer=False):
    assert os.path.isfile(checkpoint_path)
    checkpoint_dict = torch.load(checkpoint_path, map_location="cpu")
    iteration = checkpoint_dict.get("iteration", 0)
    learning_rate = checkpoint_dict.get("learning_rate", 0.)
    if (optimizer is not None and not skip_optimizer and checkpoint_dict["optimizer"] is not None):
        optimizer.load_state_dict(checkpoint_dict["optimizer"])
    elif optimizer is None and not skip_optimizer:
        # else:      Disable this line if Infer and resume checkpoint,then enable the line upper
        new_opt_dict = optimizer.state_dict()
        new_opt_dict_params = new_opt_dict["param_groups"][0]["params"]
        new_opt_dict["param_groups"] = checkpoint_dict["optimizer"]["param_groups"]
        new_opt_dict["param_groups"][0]["params"] = new_opt_dict_params
        optimizer.load_state_dict(new_opt_dict)

    saved_state_dict = checkpoint_dict["model"]
    if hasattr(model, "module"):
        state_dict = model.module.state_dict()
    else:
        state_dict = model.state_dict()

    new_state_dict = {}
    for k, v in state_dict.items():
        try:
            # assert "emb_g" not in k
            new_state_dict[k] = saved_state_dict[k]
            assert saved_state_dict[k].shape == v.shape, (
                saved_state_dict[k].shape,
                v.shape,
            )
        except Exception as e:
            print(e)
            # For upgrading from the old version
            if "ja_bert_proj" in k:
                v = torch.zeros_like(v)
                logger.warn(
                    f"Seems you are using the old version of the model, the {k} is automatically set to zero for backward compatibility"
                )
            else:
                logger.error(f"{k} is not in the checkpoint")

            new_state_dict[k] = v

    if hasattr(model, "module"):
        model.module.load_state_dict(new_state_dict, strict=False)
    else:
        model.load_state_dict(new_state_dict, strict=False)

    logger.info("Loaded checkpoint '{}' (iteration {})".format(checkpoint_path, iteration))

    return model, optimizer, learning_rate, iteration


def save_checkpoint(model, optimizer, learning_rate, iteration, checkpoint_path):
    logger.info("Saving model and optimizer state at iteration {} to {}".format(iteration, checkpoint_path))
    if hasattr(model, "module"):
        state_dict = model.module.state_dict()
    else:
        state_dict = model.state_dict()
    torch.save(
        {
            "model": state_dict,
            "iteration": iteration,
            "optimizer": optimizer.state_dict(),
            "learning_rate": learning_rate,
        },
        checkpoint_path,
    )


def summarize(
    writer,
    global_step,
    scalars={},
    histograms={},
    images={},
    audios={},
    audio_sampling_rate=22050,
):
    for k, v in scalars.items():
        writer.add_scalar(k, v, global_step)
    for k, v in histograms.items():
        writer.add_histogram(k, v, global_step)
    for k, v in images.items():
        writer.add_image(k, v, global_step, dataformats="HWC")
    for k, v in audios.items():
        writer.add_audio(k, v, global_step, audio_sampling_rate)


def latest_checkpoint_path(dir_path, regex="G_*.pth"):
    f_list = glob.glob(os.path.join(dir_path, regex))
    f_list.sort(key=lambda f: int("".join(filter(str.isdigit, f))))
    x = f_list[-1]
    return x


def plot_spectrogram_to_numpy(spectrogram):
    global MATPLOTLIB_FLAG
    if not MATPLOTLIB_FLAG:
        import matplotlib

        matplotlib.use("Agg")
        MATPLOTLIB_FLAG = True
        mpl_logger = logging.getLogger("matplotlib")
        mpl_logger.setLevel(logging.WARNING)
    import matplotlib.pylab as plt
    import numpy as np

    fig, ax = plt.subplots(figsize=(10, 2))
    im = ax.imshow(spectrogram, aspect="auto", origin="lower", interpolation="none")
    plt.colorbar(im, ax=ax)
    plt.xlabel("Frames")
    plt.ylabel("Channels")
    plt.tight_layout()

    fig.canvas.draw()
    data = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep="")
    data = data.reshape(fig.canvas.get_width_height()[::-1] + (3, ))
    plt.close()
    return data


def plot_alignment_to_numpy(alignment, info=None):
    global MATPLOTLIB_FLAG
    if not MATPLOTLIB_FLAG:
        import matplotlib

        matplotlib.use("Agg")
        MATPLOTLIB_FLAG = True
        mpl_logger = logging.getLogger("matplotlib")
        mpl_logger.setLevel(logging.WARNING)
    import matplotlib.pylab as plt
    import numpy as np

    fig, ax = plt.subplots(figsize=(6, 4))
    im = ax.imshow(alignment.transpose(), aspect="auto", origin="lower", interpolation="none")
    fig.colorbar(im, ax=ax)
    xlabel = "Decoder timestep"
    if info is not None:
        xlabel += "\n\n" + info
    plt.xlabel(xlabel)
    plt.ylabel("Encoder timestep")
    plt.tight_layout()

    fig.canvas.draw()
    data = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep="")
    data = data.reshape(fig.canvas.get_width_height()[::-1] + (3, ))
    plt.close()
    return data


def load_wav_to_torch(full_path):
    sampling_rate, data = read(full_path)
    return torch.FloatTensor(data.astype(np.float32)), sampling_rate


def load_wav_to_torch_new(full_path):
    audio_norm, sampling_rate = torchaudio.load(
        full_path, frame_offset=0, num_frames=-1, normalize=True, channels_first=True)
    audio_norm = audio_norm.mean(dim=0)
    return audio_norm, sampling_rate


def load_wav_to_torch_librosa(full_path, sr):
    audio_norm, sampling_rate = librosa.load(full_path, sr=sr, mono=True)
    return torch.FloatTensor(audio_norm.astype(np.float32)), sampling_rate


def load_filepaths_and_text(filename, split="|"):
    with open(filename, encoding="utf-8") as f:
        filepaths_and_text = [line.strip().split(split) for line in f]
    return filepaths_and_text


def get_hparams(init=True):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        default="./configs/base.json",
        help="JSON file for configuration",
    )
    parser.add_argument('--local_rank', type=int, default=0)
    parser.add_argument('--world-size', type=int, default=1)
    parser.add_argument('--port', type=int, default=10000)
    parser.add_argument("-m", "--model", type=str, required=True, help="Model name")
    parser.add_argument('--pretrain_G', type=str, default=None, help='pretrain model')
    parser.add_argument('--pretrain_D', type=str, default=None, help='pretrain model D')
    parser.add_argument('--pretrain_dur', type=str, default=None, help='pretrain model duration')

    args = parser.parse_args()
    model_dir = os.path.join("./logs", args.model)

    os.makedirs(model_dir, exist_ok=True)

    config_path = args.config
    config_save_path = os.path.join(model_dir, "config.json")
    if init:
        with open(config_path) as f:
            data = f.read()
        with open(config_save_path, "w") as f:
            f.write(data)
    else:
        with open(config_save_path) as f:
            data = f.read()
    config = json.loads(data)

    hparams = HParams(**config)
    hparams.model_dir = model_dir
    hparams.pretrain_G = args.pretrain_G
    hparams.pretrain_D = args.pretrain_D
    hparams.pretrain_dur = args.pretrain_dur
    hparams.port = args.port
    return hparams


def clean_checkpoints(path_to_models="logs/44k/", n_ckpts_to_keep=2, sort_by_time=True):
    """Freeing up space by deleting saved ckpts

    Arguments:
    path_to_models    --  Path to the model directory
    n_ckpts_to_keep   --  Number of ckpts to keep, excluding G_0.pth and D_0.pth
    sort_by_time      --  True -> chronologically delete ckpts
                          False -> lexicographically delete ckpts
    """
    import re

    ckpts_files = [f for f in os.listdir(path_to_models) if os.path.isfile(os.path.join(path_to_models, f))]

    def name_key(_f):
        return int(re.compile("._(\\d+)\\.pth").match(_f).group(1))

    def time_key(_f):
        return os.path.getmtime(os.path.join(path_to_models, _f))

    sort_key = time_key if sort_by_time else name_key

    def x_sorted(_x):
        return sorted(
            [f for f in ckpts_files if f.startswith(_x) and not f.endswith("_0.pth")],
            key=sort_key,
        )

    to_del = [
        os.path.join(path_to_models, fn)
        for fn in (x_sorted("G")[:-n_ckpts_to_keep] + x_sorted("D")[:-n_ckpts_to_keep])
    ]

    def del_info(fn):
        return logger.info(f".. Free up space by deleting ckpt {fn}")

    def del_routine(x):
        return [os.remove(x), del_info(x)]

    [del_routine(fn) for fn in to_del]


def get_hparams_from_dir(model_dir):
    config_save_path = os.path.join(model_dir, "config.json")
    with open(config_save_path, encoding="utf-8") as f:
        data = f.read()
    config = json.loads(data)

    hparams = HParams(**config)
    hparams.model_dir = model_dir
    return hparams


def get_hparams_from_file(config_path):
    with open(config_path, encoding="utf-8") as f:
        data = f.read()
    config = json.loads(data)

    hparams = HParams(**config)
    return hparams


def check_git_hash(model_dir):
    source_dir = os.path.dirname(os.path.realpath(__file__))
    if not os.path.exists(os.path.join(source_dir, ".git")):
        logger.warn(
            "{} is not a git repository, therefore hash value comparison will be ignored.".format(source_dir))
        return

    cur_hash = subprocess.getoutput("git rev-parse HEAD")

    path = os.path.join(model_dir, "githash")
    if os.path.exists(path):
        saved_hash = open(path).read()
        if saved_hash != cur_hash:
            logger.warn(
                "git hash values are different. {}(saved) != {}(current)".format(
                    saved_hash[:8], cur_hash[:8]))
    else:
        open(path, "w").write(cur_hash)


def get_logger(model_dir, filename="train.log"):
    global logger
    logger = logging.getLogger(os.path.basename(model_dir))
    logger.setLevel(logging.DEBUG)

    formatter = logging.Formatter("%(asctime)s\t%(name)s\t%(levelname)s\t%(message)s")
    if not os.path.exists(model_dir):
        os.makedirs(model_dir, exist_ok=True)
    h = logging.FileHandler(os.path.join(model_dir, filename))
    h.setLevel(logging.DEBUG)
    h.setFormatter(formatter)
    logger.addHandler(h)
    return logger


class HParams:

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if type(v) == dict:
                v = HParams(**v)
            self[k] = v

    def keys(self):
        return self.__dict__.keys()

    def items(self):
        return self.__dict__.items()

    def values(self):
        return self.__dict__.values()

    def __len__(self):
        return len(self.__dict__)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def __contains__(self, key):
        return key in self.__dict__

    def __repr__(self):
        return self.__dict__.__repr__()



# https://github.com/myshell-ai/MeloTTS/pull/117

import re
import pytest
import torch
from meloplus.text.thai import (
    normalize,
    word_tokenize,
    thai_text_to_phonemes,
    text_normalize,
    g2p,
    get_bert_feature,
)
from meloplus.text.korean import (
    text_normalize as k_text_normalize,
    get_bert_feature as k_get_bert_feature,
    g2p as k_g2p,
)


def test_normalize():
    text = "   ข้อความ ภาษา ไทย 123 ABC   "
    normalized_text = normalize(text)
    assert normalized_text == "ข้อความ ภาษา ไทย หนึ่งร้อยยี่สิบสาม เอบีซี"


# def test_word_tokenize():
#     text = "ฉันเข้าใจคุณค่าของงานของฉันและความหมายของสิ่งที่ฟอนเทนทำเพื่อคนทั่วไปเป็นอย่างดี"
#     tokenized_text = word_tokenize(text, engine="newmm")
#     assert tokenized_text == ['ฉัน', 'เข้าใจ', 'คุณค่า', 'ของ', 'งาน', 'ของ', 'ฉัน', 'และ', 'ความหมาย', 'ของ', 'สิ่ง', 'ที่', 'ฟอน', 'เท', 'น', 'ทำ', 'เพื่อ', 'คน', 'ทั่วไป', 'เป็น', 'อย่าง', 'ดี']

# def test_thai_text_to_phonemes():
#     text = "สวัสดีครับ"
#     phonemes = thai_text_to_phonemes(text)
#     assert phonemes == "s a ˨˩ . w a t̚ ˨˩ . d iː ˧ kʰ r a p̚ ˦˥"

# def test_g2p():
#     text = "กงล้อ"
#     normalized_text = text_normalize(text)
#     phones, tones, word2ph = g2p(normalized_text)

#     print(f"Phones: {phones}")
#     print(f"Tones: {tones}")
#     print(f"Word2ph: {word2ph}")

#     expected_phones = ['_', 't͡ɕʰ', 'a', 'n', '', 'r', 'a', 'k̚', '', 'm', 'ɯa̯', 'ŋ', '', 'tʰ', 'aj', '', '.', 'j', 'a', '', '.', '_']
#     expected_tones = [0, 0, 0, 4, 0, 0, 0, 5, 0, 0, 0, 3, 0, 0, 3, 0, 0, 0, 5, 0]
#     expected_word2ph = [1, 0, 1, 1, 1]

#     assert phones == expected_phones
#     assert tones == expected_tones
#     assert word2ph == expected_word2ph

#     # Additional test case
#     text = "สวัสดี ประเทศไทย"
#     normalized_text = text_normalize(text)
#     phones, tones, word2ph = g2p(normalized_text)

#     print(f"Phones: {phones}")
#     print(f"Tones: {tones}")
#     print(f"Word2ph: {word2ph}")

#     expected_phones = ['_', 's', 'a', 'w', 'a', 't̚', '', 'd', 'iː', '', 'p', 'r', 'a', '', 'tʰ', 'eː', 't̚', '', 'tʰ', 'aj', '', '.', 'j', 'a', '', '.', '_']
#     expected_tones = [0, 0, 0, 0, 4, 2, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 3, 0, 0, 0, 5, 0]
#     expected_word2ph = [1, 6, 10, 14, 18, 22, 26, 1]

#     assert phones == expected_phones
#     assert tones == expected_tones
#     assert word2ph == expected_word2ph

# def test_get_bert_feature():
#     text = "ฉันเข้าใจคุณค่าของงานของฉันและความหมายของสิ่งที่ฟอนเทนทำเพื่อคนทั่วไปเป็นอย่างดี"
#     normalized_text = text_normalize(text)
#     phones, tones, word2ph = g2p(normalized_text)

#     bert_features = get_bert_feature(normalized_text, word2ph, device='cpu')

#     assert isinstance(bert_features, torch.Tensor), "bert_features should be a torch.Tensor"
#     assert bert_features.shape[0] == 768, f"Expected bert_features.shape[0] to be 768, but got {bert_features.shape[0]}"

#     # Modify the assertion to check the number of phones instead of the length of word2ph
#     num_phones = sum(word2ph)
#     assert bert_features.shape[1] == num_phones, f"Expected bert_features.shape[1] to be {num_phones}, but got {bert_features.shape[1]}"

#     # Additional assertions to check the values of bert_features
#     assert not torch.isnan(bert_features).any(), "bert_features should not contain any NaN values"
#     assert not torch.isinf(bert_features).any(), "bert_features should not contain any infinity values"


def extract_word_and_phonemes(line):
    parts = line.strip().split("\t")
    if len(parts) == 2:
        word, phonemes = parts
        return word, phonemes.split()
    return None


def test_g2p():
    # Test case for the word "กงล้อ"
    text = "กงล้อ"
    normalized_text = text_normalize(text)
    phones, tones, word2ph = g2p(normalized_text)

    # Expected output based on the wiktionary entry
    expected_phones = ['_', 'k', 'o', 'ŋ', 'l', 'ɔː', '_']
    expected_tones = [1, 2, 2, 2, 3, 3, 1]
    expected_word2ph = [1, 3, 2, 1]

    # Compare the actual output with the expected output
    assert phones == expected_phones
    assert tones == expected_tones
    assert word2ph == expected_word2ph


def test_get_bert_feature_thai():
    text = "กงล้อ"
    normalized_text = text_normalize(text)
    phones, tones, word2ph = g2p(normalized_text)
    bert_features = get_bert_feature(normalized_text, word2ph, device='cpu')
    assert isinstance(bert_features, torch.Tensor), "bert_features should be a torch.Tensor"
    assert bert_features.shape[
        0] == 768, f"Expected bert_features.shape[0] to be 768, but got {bert_features.shape[0]}"

    # Modify the assertion to check the number of phones, excluding special characters
    num_phones = sum(word2ph)
    assert bert_features.shape[
        1] == num_phones, f"Expected bert_features.shape[1] to be {num_phones}, but got {bert_features.shape[1]}"

    assert not torch.isnan(bert_features).any(), "bert_features should not contain any NaN values"
    assert not torch.isinf(bert_features).any(), "bert_features should not contain any infinity values"


# Compare with Korean
def test_get_bert_feature_korean():
    text = "저는 제 일의 가치와 의미를 잘 알고 있습니다. 앞으로도 저는 제 일에 자부심을 갖고 살아갈 것입니다."
    normalized_text = k_text_normalize(text)
    phones, tones, word2ph = k_g2p(normalized_text)

    bert_features = k_get_bert_feature(normalized_text, word2ph, device='cpu')

    assert isinstance(bert_features, torch.Tensor), "bert_features should be a torch.Tensor"
    assert bert_features.shape[
        0] == 768, f"Expected bert_features.shape[0] to be 768, but got {bert_features.shape[0]}"

    # Modify the assertion to check the number of phones instead of the length of word2ph
    num_phones = sum(word2ph)
    assert bert_features.shape[
        1] == num_phones, f"Expected bert_features.shape[1] to be {num_phones}, but got {bert_features.shape[1]}"

    # Additional assertions to check the values of bert_features
    assert not torch.isnan(bert_features).any(), "bert_features should not contain any NaN values"
    assert not torch.isinf(bert_features).any(), "bert_features should not contain any infinity values"



import os
import torch
from cached_path import cached_path
from . import utils

DOWNLOAD_CKPT_URLS = {
    'EN': 'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/basespeakers/EN/checkpoint.pth',
    'EN_V2': 'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/basespeakers/EN_V2/checkpoint.pth',
    'FR': 'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/basespeakers/FR/checkpoint.pth',
    'JP': 'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/basespeakers/JP/checkpoint.pth',
    'ES': 'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/basespeakers/ES/checkpoint.pth',
    'ZH': 'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/basespeakers/ZH/checkpoint.pth',
    'KR': 'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/basespeakers/KR/checkpoint.pth',
}

DOWNLOAD_CONFIG_URLS = {
    'EN': 'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/basespeakers/EN/config.json',
    'EN_V2': 'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/basespeakers/EN_V2/config.json',
    'FR': 'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/basespeakers/FR/config.json',
    'JP': 'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/basespeakers/JP/config.json',
    'ES': 'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/basespeakers/ES/config.json',
    'ZH': 'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/basespeakers/ZH/config.json',
    'KR': 'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/basespeakers/KR/config.json',
}

PRETRAINED_MODELS = {
    'G.pth': 'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/basespeakers/pretrained/G.pth',
    'D.pth': 'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/basespeakers/pretrained/D.pth',
    'DUR.pth': 'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/basespeakers/pretrained/DUR.pth',
}


def load_or_download_config(locale, use_hf=True, config_path=None):
    if config_path is None:
        raise ValueError("config_path must be provided")
    return utils.get_hparams_from_file(config_path)


def load_or_download_model(locale, device, use_hf=True, ckpt_path=None):
    if ckpt_path is None:
        raise ValueError("ckpt_path must be provided")
    return torch.load(ckpt_path, map_location=device)


def load_pretrain_model():
    raise NotImplementedError("This function is not supported. Please provide model paths directly.")



import re

# List of (regular expression, replacement) pairs for abbreviations in french:
abbreviations_fr = [(re.compile("\\b%s\\." % x[0], re.IGNORECASE), x[1]) for x in [
    ("M", "monsieur"),
    ("Mlle", "mademoiselle"),
    ("Mlles", "mesdemoiselles"),
    ("Mme", "Madame"),
    ("Mmes", "Mesdames"),
    ("N.B", "nota bene"),
    ("M", "monsieur"),
    ("p.c.q", "parce que"),
    ("Pr", "professeur"),
    ("qqch", "quelque chose"),
    ("rdv", "rendez-vous"),
    ("max", "maximum"),
    ("min", "minimum"),
    ("no", "numéro"),
    ("adr", "adresse"),
    ("dr", "docteur"),
    ("st", "saint"),
    ("co", "companie"),
    ("jr", "junior"),
    ("sgt", "sergent"),
    ("capt", "capitain"),
    ("col", "colonel"),
    ("av", "avenue"),
    ("av. J.-C", "avant Jésus-Christ"),
    ("apr. J.-C", "après Jésus-Christ"),
    ("art", "article"),
    ("boul", "boulevard"),
    ("c.-à-d", "c’est-à-dire"),
    ("etc", "et cetera"),
    ("ex", "exemple"),
    ("excl", "exclusivement"),
    ("boul", "boulevard"),
]] + [(re.compile("\\b%s" % x[0]), x[1]) for x in [
    ("Mlle", "mademoiselle"),
    ("Mlles", "mesdemoiselles"),
    ("Mme", "Madame"),
    ("Mmes", "Mesdames"),
]]



# https://github.com/g-hano/MeloTTS/blob/main/melo/text/turkish.py

import re
from transformers import AutoTokenizer
from . import symbols


def distribute_phone(n_phone, n_word):
    phones_per_word = [0] * n_word
    for task in range(n_phone):
        min_tasks = min(phones_per_word)
        min_index = phones_per_word.index(min_tasks)
        phones_per_word[min_index] += 1
    return phones_per_word


def text_normalize(text):
    # Basic Turkish text normalization
    # Convert to lowercase while preserving Turkish characters
    text = text.replace("I", "ı").lower()
    text = text.lower()
    # Replace multiple spaces with single space
    text = re.sub(r'\s+', ' ', text)
    # Remove unnecessary punctuation
    text = re.sub(r'[^\w\s.,!?;:əğıöüşçİĞÖÜŞÇ]', '', text)
    return text.strip()


def post_replace_ph(ph):
    rep_map = {
        "：": ",",
        "；": ",",
        "，": ",",
        "。": ".",
        "！": "!",
        "？": "?",
        "\n": ".",
        "·": ",",
        "、": ",",
        "...": "…"
    }
    if ph in rep_map.keys():
        ph = rep_map[ph]
    if ph in symbols:
        return ph
    if ph not in symbols:
        ph = "UNK"
    return ph


def refine_ph(phn):
    tone = 0
    if re.search(r"\d$", phn):
        tone = int(phn[-1]) + 1
        phn = phn[:-1]
    return phn.lower(), tone


def tr_to_ipa(text):
    """
    Convert Turkish text to IPA
    This is a basic implementation - you might want to expand this based on Turkish phonology rules
    """
    tr_to_ipa_dict = {
        'a': 'a',
        'e': 'e',
        'ı': 'ɯ',
        'i': 'i',
        'o': 'o',
        'ö': 'ø',
        'u': 'u',
        'ü': 'y',
        'b': 'b',
        'c': 'dʒ',
        'ç': 'tʃ',
        'd': 'd',
        'f': 'f',
        'g': 'ɡ',
        'ğ': 'ː',
        'h': 'h',
        'j': 'ʒ',
        'k': 'k',
        'l': 'l',
        'm': 'm',
        'n': 'n',
        'p': 'p',
        'r': 'r',
        's': 's',
        'ş': 'ʃ',
        't': 't',
        'v': 'v',
        'y': 'j',
        'z': 'z'
    }
    return ''.join(tr_to_ipa_dict.get(char, char) for char in text.lower())


# Initialize the Turkish BERT tokenizer
model_id = 'ytu-ce-cosmos/turkish-base-bert-uncased'
tokenizer = AutoTokenizer.from_pretrained(model_id)


def g2p(text, pad_start_end=True, tokenized=None):
    if tokenized is None:
        tokenized = tokenizer.tokenize(text)

    phs = []
    ph_groups = []
    for t in tokenized:
        if not t.startswith("##"):  # Note: Turkish BERT uses ## for subwords
            ph_groups.append([t])
        else:
            ph_groups[-1].append(t.replace("##", ""))

    phones = []
    tones = []
    word2ph = []

    for group in ph_groups:
        w = "".join(group)
        phone_len = 0
        word_len = len(group)
        if w == '[UNK]':
            phone_list = ['UNK']
        else:
            phone_list = list(filter(lambda p: p != " ", tr_to_ipa(w)))

        for ph in phone_list:
            phones.append(ph)
            tones.append(0)  # Turkish is not a tonal language
            phone_len += 1
        aaa = distribute_phone(phone_len, word_len)
        word2ph += aaa

    if pad_start_end:
        phones = ["_"] + phones + ["_"]
        tones = [0] + tones + [0]
        word2ph = [1] + word2ph + [1]
    return phones, tones, word2ph


def get_bert_feature(text, word2ph, device=None):
    from text import turkish_bert
    return turkish_bert.get_bert_feature(text, word2ph, device=device)


if __name__ == "__main__":
    text = "Merhaba, nasılsın? Ben iyiyim."
    text = text_normalize(text)
    print(text)
    phones, tones, word2ph = g2p(text)
    bert = get_bert_feature(text, word2ph)
    print(phones)
    print(len(phones), tones, sum(word2ph), bert.shape)



# https://github.com/myshell-ai/MeloTTS/pull/117
import torch
from transformers import AutoTokenizer, AutoModelForMaskedLM
import sys

models = {}
tokenizers = {}


def get_bert_feature(text, word2ph, device=None, model_id='clicknext/phayathaibert'):
    global model
    global tokenizer

    if (sys.platform == "darwin" and torch.backends.mps.is_available() and device == "cpu"):
        device = "mps"
    if not device:
        device = "cuda"
    if model_id not in models:
        model = AutoModelForMaskedLM.from_pretrained(model_id).to(device)
        models[model_id] = model
        tokenizer = AutoTokenizer.from_pretrained(model_id)
        tokenizers[model_id] = tokenizer
    else:
        model = models[model_id]
        tokenizer = tokenizers[model_id]

    with torch.no_grad():
        inputs = tokenizer(text, return_tensors="pt")
        tokenized = tokenizer.tokenize(text)
        for i in inputs:
            inputs[i] = inputs[i].to(device)
        res = model(**inputs, output_hidden_states=True)
        res = torch.cat(res["hidden_states"][-3:-2], -1)[0].cpu()

    try:
        assert inputs["input_ids"].shape[-1] == len(
            word2ph
        ), f"Input IDs length: {inputs['input_ids'].shape[-1]} / Word2ph length: {len(word2ph)}, Tokenized: {tokenized}, Inputs: {inputs}, Word2ph: {word2ph}"
    except AssertionError as e:
        print(f"Warning: Mismatch in input lengths. Details: {e}")

    word2phone = word2ph
    phone_level_feature = []
    for i in range(len(word2phone)):
        repeat_feature = res[i].repeat(word2phone[i], 1)
        phone_level_feature.append(repeat_feature)

    phone_level_feature = torch.cat(phone_level_feature, dim=0)

    return phone_level_feature.T



# https://github.com/myshell-ai/MeloTTS/pull/117

import re
import unicodedata
from transformers import AutoTokenizer
from . import punctuation, symbols, pu_symbols
from num2words import num2words
from pythainlp.tokenize import word_tokenize
from pythainlp.transliterate import romanize
from pythainlp.util import normalize as thai_normalize
from pythainlp.util import thai_to_eng, eng_to_thai
from meloplus.text.thai_dictionary import english_dictionary, etc_dictionary
from collections import defaultdict
import logging
import os

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def normalize_with_dictionary(text, dic):
    if any(key in text for key in dic.keys()):
        pattern = re.compile("|".join(re.escape(key) for key in dic.keys()))
        return pattern.sub(lambda x: dic[x.group()], text)
    return text


def normalize(text):
    text = text.strip()
    text = thai_normalize(text)
    text = normalize_with_dictionary(text, etc_dictionary)
    text = re.sub(r"\d+", lambda x: num2words(int(x.group()), lang="th"), text)
    text = normalize_english(text)
    text = text.lower()
    return text


def normalize_english(text):

    def fn(m):
        word = m.group()
        if word.upper() in english_dictionary:
            return english_dictionary[word.upper()]
        return "".join(english_dictionary.get(char.upper(), char) for char in word)

    text = re.sub(r"([A-Za-z]+)", fn, text)
    return text


# Load the Thai G2P dictionary
thai_g2p_dict = defaultdict(list)

dict_path = os.path.join(os.path.dirname(__file__), "wiktionary-23-7-2022-clean.tsv")
with open(dict_path, encoding="utf-8") as f:
    for line in f:
        word, phonemes = line.strip().split("\t")
        thai_g2p_dict[word].append(phonemes.split())
        thai_g2p_dict["ะ"] = ["a"]


def map_word_to_phonemes(word):
    logger.debug(f"Looking up word: {word}")

    # First, try to find the whole word in the dictionary
    phonemes_list = thai_g2p_dict.get(word)
    if phonemes_list:
        logger.debug(f"Found whole word {word} in dictionary")
        return " ".join(phonemes_list[0])

    # If not found, try to split the word
    subwords = word_tokenize(word, engine="newmm")

    if len(subwords) > 1:
        logger.debug(f"Word {word} split into subwords: {subwords}")
        # If the word can be split, recursively process each subword
        return " . ".join(map_word_to_phonemes(subword) for subword in subwords)
    else:
        logger.debug(f"Word {word} cannot be split, processing character by character")
        return map_partial_word(word)


def map_partial_word(word):
    if not word:
        return ""

    logger.debug(f"Mapping partial word: {word}")

    # Handle Thanthakhat (์) character
    if len(word) > 1 and word[1] == '์':
        logger.debug(f"Found Thanthakhat, skipping {word[:2]}")
        return map_partial_word(word[2:])

    # Handle vowels and tone marks
    if word[0] in thai_vowels or word[0] in thai_tone_marks:
        phoneme = thai_g2p_dict.get(word[0], [word[0]])[0]
        return phoneme + " " + map_partial_word(word[1:])

    # Try to find the longest matching prefix
    for i in range(len(word), 0, -1):
        prefix = word[:i]
        phonemes_list = thai_g2p_dict.get(prefix)
        if phonemes_list:
            logger.debug(f"Found matching prefix: {prefix}")
            return " ".join(phonemes_list[0]) + " " + map_partial_word(word[i:])

    # If no match found, return the first character and continue with the rest
    logger.debug(f"No match found for {word[0]}, continuing with rest")
    return word[0] + " " + map_partial_word(word[1:])


# Comprehensive mapping of Thai characters to their phonetic representations
thai_char_to_phoneme = {
    # Consonants
    'ก': 'k',
    'ข': 'kʰ',
    'ฃ': 'kʰ',
    'ค': 'kʰ',
    'ฅ': 'kʰ',
    'ฆ': 'kʰ',
    'ง': 'ŋ',
    'จ': 't͡ɕ',
    'ฉ': 't͡ɕʰ',
    'ช': 't͡ɕʰ',
    'ซ': 's',
    'ฌ': 't͡ɕʰ',
    'ญ': 'j',
    'ฎ': 'd',
    'ฏ': 't',
    'ฐ': 'tʰ',
    'ฑ': 'tʰ',
    'ฒ': 'tʰ',
    'ณ': 'n',
    'ด': 'd',
    'ต': 't',
    'ถ': 'tʰ',
    'ท': 'tʰ',
    'ธ': 'tʰ',
    'น': 'n',
    'บ': 'b',
    'ป': 'p',
    'ผ': 'pʰ',
    'ฝ': 'f',
    'พ': 'pʰ',
    'ฟ': 'f',
    'ภ': 'pʰ',
    'ม': 'm',
    'ย': 'j',
    'ร': 'r',
    'ล': 'l',
    'ว': 'w',
    'ศ': 's',
    'ษ': 's',
    'ส': 's',
    'ห': 'h',
    'ฬ': 'l',
    'อ': 'ʔ',
    'ฮ': 'h',

    # Vowels
    'ะ': 'a',
    'ั': 'a',
    'า': 'aː',
    'ำ': 'am',
    'ิ': 'i',
    'ี': 'iː',
    'ึ': 'ɯ',
    'ื': 'ɯː',
    'ุ': 'u',
    'ู': 'uː',
    'เ': 'eː',
    'แ': 'ɛː',
    'โ': 'oː',
    'ใ': 'aj',
    'ไ': 'aj',
    '็': '',  # Short vowel marker
    'ๆ': '',  # Repetition marker

    # Tone marks
    '่': '˨˩',  # Low tone
    '้': '˦˥',  # Rising tone
    '๊': '˥˩',  # Falling tone
    '๋': '˧',  # High tone

    # Special characters
    '์': '',  # Thanthakhat (cancels sound of preceding consonant)
}


def map_remaining_thai_chars(phones):
    mapped_phones = []
    for phone in phones:
        if phone in thai_char_to_phoneme:
            mapped_phones.append(thai_char_to_phoneme[phone])
        else:
            mapped_phones.append(phone)
    return mapped_phones


def thai_text_to_phonemes(text):
    text = normalize(text)
    words = word_tokenize(text, engine="newmm")
    logger.debug(f"word_tokenize output: {words}")
    phonemes = []
    for word in words:
        word_phonemes = map_word_to_phonemes(word)
        phonemes.extend(word_phonemes.split())

    # Map any remaining Thai characters
    mapped_phonemes = map_remaining_thai_chars(phonemes)

    return " ".join(mapped_phonemes)


# Define Thai vowels, tone marks, and special characters
thai_vowels = set("ะัาำิีึืุูเแโใไฤฦ็")
thai_tone_marks = set("่้๊๋")
thai_special_chars = set("์ๆฯ๎๏")  # Thanthakhat, Maiyamok, Paitai, and Phinthu

# Update the thai_g2p_dict with proper mappings for vowels and tone marks
thai_g2p_dict.update({
    'โ': ['o'],
    'ใ': ['aj'],
    'ไ': ['aj'],
    'แ': ['ɛː'],
    'เ': ['eː'],
    'ฤ': ['rɯ'],
    'ฦ': ['lɯ'],
    '็': ['ː'],  # Mai Taikhu (used to shorten a vowel)
    '่': ['˨˩'],  # Low tone
    '้': ['˦˥'],  # Rising tone
    '๊': ['˥˩'],  # Falling tone
    '๋': ['˧'],  # High tone
    '์': [''],  # Thanthakhat (cancels the sound of the syllable)
    'ๆ': [''],  # Maiyamok (repetition mark)
    'ฯ': [''],  # Paitai (abbreviation mark)
    '๎': [''],  # Phinthu (used to indicate a silent consonant)
    '๏': [''],  # Angkhankhu (used to mark the end of a paragraph or section)
})


def text_normalize(text):
    text = normalize(text)
    return text


def distribute_phone(n_phone, n_word):
    phones_per_word = [0] * n_word
    for task in range(n_phone):
        min_tasks = min(phones_per_word)
        min_index = phones_per_word.index(min_tasks)
        phones_per_word[min_index] += 1
    return phones_per_word


model_id = 'clicknext/phayathaibert'
tokenizer = AutoTokenizer.from_pretrained(model_id)

tone_map = {
    "˧": 2,  # Mid tone
    "˨˩": 1,  # Low tone
    "˦˥": 3,  # Rising tone
    "˩˩˦": 4,  # Falling tone
    "˥˩": 5,  # High tone
}

# def extract_tones(phs):
#     tones = []
#     tone_value = 2  # Default tone value when no tone symbol is found
#     last_item = phs[-1]

#     for ph in phs:
#         if ph in tone_map:
#             tone_value = tone_map[ph]

#     # Assign the tone value to each phoneme excluding the tone marker if there is one
#     if last_item in tone_map:
#         tones = [tone_value] * (len(phs) - 1)
#     else:
#         tones = [tone_value] * (len(phs))

#     print("=======> PHS: ",  phs)
#     print("=======> TONES: ",  tones)

#     return tones


def extract_tones(phs):
    phonemes = []
    tones = []
    current_tone = 2  # Default mid tone
    for ph in phs.split():
        if ph in tone_map:
            current_tone = tone_map[ph]
        elif ph == '_':
            phonemes.append(ph)
            tones.append(0)  # Zero tone for underscore
        else:
            phonemes.append(ph)
            tones.append(current_tone)
    return phonemes, tones


# def g2p_bert(norm_text, pad_start_end=True):
#     tokenized = tokenizer.tokenize(norm_text)
#     print("tokenized text")
#     phs = []
#     tones = []
#     word2ph = []

#     ph_groups = []
#     for t in tokenized:
#         if t.startswith("▁"):
#             ph_groups.append([t])
#         else:
#             ph_groups[-1].append(t)

#     for group in ph_groups:
#         word = "".join(group).replace("▁", "")
#         phonemes = thai_text_to_phonemes(word)
#         phoneme_groups = phonemes.split(".")
#         phoneme_groups = list(filter(str.strip, phoneme_groups))

#         word_phonemes = []
#         word_tones = []
#         word_phones_count = []

#         for p_group in phoneme_groups:
#             group_phonemes = [ph for ph in p_group.split() if ph not in tone_map]
#             group_tones = extract_tones(p_group.split())

#             word_phonemes.extend(group_phonemes)
#             word_tones.extend(group_tones)
#             word_phones_count.append(len(group_phonemes))

#         phs.extend(word_phonemes)
#         tones.extend(word_tones)
#         word2ph.extend(word_phones_count)

#     if pad_start_end:
#         phs = ["_"] + phs + ["_"]
#         tones = [1] + tones + [1]
#         word2ph = [1] + word2ph + [1]

#     print(f"Final phs: {phs}")
#     print(f"Final tones: {tones}")
#     print(f"Final word2ph: {word2ph}")

#     assert len(word2ph) == len(tokenized) + 2

#     return phs, tones, word2ph


def g2p_og(norm_text, pad_start_end=True):
    tokenized = tokenizer.tokenize(norm_text)
    phs = []
    tones = []
    word2ph = []

    print("tokenized", tokenized)

    for word in tokenized:
        # if word == "▁":
        #     word2ph.append(1)
        #     continue

        if word.startswith("▁"):
            word = word.replace("▁", "")

        phonemes = thai_text_to_phonemes(word)
        phoneme_groups = phonemes.split(".")
        # Keep only non-empty groups for cases with a trailing dot
        # i.e 'b ɤː ˧ . tʰ oː ˧ . r a ˦˥ .'
        phoneme_groups = [group for group in phoneme_groups if group.strip()]

        word_phonemes = []
        word_tones = []

        for group in phoneme_groups:
            group_phonemes = [ph for ph in group.split() if ph not in tone_map]
            group_tones = extract_tones(group.split())
            word_phonemes.extend(group_phonemes)
            word_tones.extend(group_tones)

        phs.extend(word_phonemes)
        tones.extend(word_tones)
        word2ph.append(len(word_phonemes))

    if pad_start_end:
        phs = ["_"] + phs + ["_"]
        tones = [1] + tones + [1]
        word2ph = [1] + word2ph + [1]

    print(f"Final phs: {phs}")
    print(f"Final tones: {tones}")
    print(f"Final word2ph: {word2ph}")

    # assert len(word2ph) == len([t for t in tokenized if t != "▁"]) + 2

    return phs, tones, word2ph


def g2p_no_undersscores(norm_text, pad_start_end=True):
    tokenized = tokenizer.tokenize(norm_text)
    # print("The tokenized text", tokenized)
    phs = []
    tones = []
    ph_groups = []

    # Group tokens
    for t in tokenized:
        if not t.startswith("#"):
            ph_groups.append([t])
        else:
            ph_groups[-1].append(t.replace("#", ""))

    word2ph = []
    for group in ph_groups:
        text = "".join(group)

        # Special token handling
        if text == '[UNK]':
            phs += ['_']
            tones += [0]
            word2ph += [1]
            continue
        elif text in punctuation:
            phs += [text]
            tones += [0]
            word2ph += [1]
            continue

        # Phoneme conversion for grouped text
        phonemes = thai_text_to_phonemes(text)
        phoneme_groups = phonemes.split(".")
        phoneme_groups = list(filter(str.strip, phoneme_groups))

        word_phonemes = []
        word_tones = []
        for p_group in phoneme_groups:
            group_phonemes = [ph for ph in p_group.split() if ph not in tone_map]
            group_tones = extract_tones(p_group.split())
            word_phonemes.extend(group_phonemes)
            word_tones.extend(group_tones)

        phone_len = len(word_phonemes)
        word_len = len(group)

        aaa = distribute_phone(phone_len, word_len)
        assert len(aaa) == word_len
        word2ph += aaa

        phs += word_phonemes
        tones += word_tones

    if pad_start_end:
        phs = ["_"] + phs + ["_"]
        tones = [0] + tones + [0]
        word2ph = [1] + word2ph + [1]

    # print(f"Final phs: {phs}")
    # print(f"Final tones: {tones}")
    # print(f"Final word2ph: {word2ph}")

    assert len(word2ph) == len(tokenized) + 2
    return phs, tones, word2ph


def g2p(norm_text, pad_start_end=True):
    tokenized = tokenizer.tokenize(norm_text)
    phs = []
    tones = []
    ph_groups = []

    # Group tokens
    for t in tokenized:
        if t == '▁':
            ph_groups.append([t])  # Add '▁' as its own group
        else:
            if not ph_groups or ph_groups[-1] == ['▁']:
                ph_groups.append([t])
            else:
                ph_groups[-1].append(t)

    word2ph = []
    for group in ph_groups:
        text = "".join(group)
        if text == '▁':
            phs += ['_']
            tones += [0]  # Zero tone for underscore
            word2ph += [1]
            continue
        elif text == '[UNK]':
            phs += ['_']
            tones += [0]  # Zero tone for unknown token
            word2ph += [1]
            continue
        elif text in punctuation:
            phs += [text]
            tones += [0]  # Zero tone for punctuation
            word2ph += [1]
            continue

        # Phoneme conversion for grouped text
        phonemes = thai_text_to_phonemes(text)
        phoneme_groups = phonemes.split(".")
        phoneme_groups = list(filter(str.strip, phoneme_groups))

        word_phonemes = []
        word_tones = []
        for p_group in phoneme_groups:
            group_phonemes, group_tones = extract_tones(p_group)
            word_phonemes.extend(group_phonemes)
            word_tones.extend(group_tones)

        phone_len = len(word_phonemes)
        word_len = len(group)
        aaa = distribute_phone(phone_len, word_len)
        assert len(aaa) == word_len
        word2ph += aaa
        phs += word_phonemes
        tones += word_tones

    if pad_start_end:
        phs = ["_"] + phs + ["_"]
        tones = [0] + tones + [0]  # Zero tone for start/end padding
        word2ph = [1] + word2ph + [1]

    # assert len(phs) == len(tones)
    # assert len(phs) == sum(word2ph)

    return phs, tones, word2ph


def get_bert_feature(text, word2ph, device='cuda', model_id='clicknext/phayathaibert'):
    from . import thai_bert
    return thai_bert.get_bert_feature(text, word2ph, device=device, model_id=model_id)


if __name__ == "__main__":
    try:
        from text.symbols import symbols
        text = "ฉันเข้าใจคุณค่าของงานของฉันและความหมายของสิ่งที่ฟอนเทนทำเพื่อคนทั่วไปเป็นอย่างดี ฉันจะใช้ชีวิตอย่างภาคภูมิใจในงานของฉันต่อไป"

        text = text_normalize(text)
        phones, tones, word2ph = g2p(text)
        bert = get_bert_feature(text, word2ph, device='cuda', model_id=model_id)

        new_symbols = []
        for ph in phones:
            if ph not in symbols and ph not in new_symbols:
                new_symbols.append(ph)
                print('update!, now symbols:')
                print(new_symbols)
                with open('thai_symbol.txt', 'w') as f:
                    f.write(f'{new_symbols}')

    except Exception as e:
        print(f"An error occurred: {e}")



import re

# List of (regular expression, replacement) pairs for abbreviations in english:
abbreviations_en = [(re.compile("\\b%s\\." % x[0], re.IGNORECASE), x[1]) for x in [
    ("mrs", "misess"),
    ("mr", "mister"),
    ("dr", "doctor"),
    ("st", "saint"),
    ("co", "company"),
    ("jr", "junior"),
    ("maj", "major"),
    ("gen", "general"),
    ("drs", "doctors"),
    ("rev", "reverend"),
    ("lt", "lieutenant"),
    ("hon", "honorable"),
    ("sgt", "sergeant"),
    ("capt", "captain"),
    ("esq", "esquire"),
    ("ltd", "limited"),
    ("col", "colonel"),
    ("ft", "fort"),
]]


def expand_abbreviations(text, lang="en"):
    if lang == "en":
        _abbreviations = abbreviations_en
    else:
        raise NotImplementedError()
    for regex, replacement in _abbreviations:
        text = re.sub(regex, replacement, text)
    return text
